#include "td_main.h"
#include <raylib.h>
#include <stdio.h>

typedef struct ParserState
{
  char *input;
  int position;
  char nextToken[256];
} ParserState;

int ParserStateGetLineNumber(ParserState *state)
{
  int lineNumber = 1;
  for (int i = 0; i < state->position; i++)
  {
    if (state->input[i] == '\n')
    {
      lineNumber++;
    }
  }
  return lineNumber;
}

void ParserStateSkipWhiteSpaces(ParserState *state)
{
  char *input = state->input;
  int pos = state->position;
  int skipped = 1;
  while (skipped)
  {
    skipped = 0;
    if (input[pos] == '-' && input[pos + 1] == '-')
    {
      skipped = 1;
      // skip comments
      while (input[pos] != 0 && input[pos] != '\n')
      {
        pos++;
      }
    }
  
    // skip white spaces and ignore colons
    while (input[pos] != 0 && (input[pos] <= ' ' || input[pos] == ':'))
    {
      skipped = 1;
      pos++;
    }

    // repeat until no more white spaces or comments
  }
  state->position = pos;
}

int ParserStateReadNextToken(ParserState *state)
{
  ParserStateSkipWhiteSpaces(state);

  int i = 0, pos = state->position;
  char *input = state->input;

  // read token
  while (input[pos] != 0 && input[pos] > ' ' && input[pos] != ':' && i < 256)
  {
    state->nextToken[i] = input[pos];
    pos++;
    i++;
  }
  state->position = pos;

  if (i == 0 || i == 256)
  {
    state->nextToken[0] = 0;
    return 0;
  }
  // terminate the token
  state->nextToken[i] = 0;
  return 1;
}

int ParserStateReadNextInt(ParserState *state, int *value)
{
  if (!ParserStateReadNextToken(state))
  {
    return 0;
  }
  // check if the token is a number
  int isSigned = state->nextToken[0] == '-' || state->nextToken[0] == '+';
  for (int i = isSigned; state->nextToken[i] != 0; i++)
  {
    if (state->nextToken[i] < '0' || state->nextToken[i] > '9')
    {
      return 0;
    }
  }
  *value = TextToInteger(state->nextToken);
  return 1;
}

typedef struct ParsedGameData
{
  const char *parseError;
  Level levels[32];
  int lastLevelIndex;
  TowerTypeConfig towerTypes[TOWER_TYPE_COUNT];
  EnemyClassConfig enemyClasses[8];
} ParsedGameData;

typedef enum TryReadResult
{
  TryReadResult_NoMatch,
  TryReadResult_Error,
  TryReadResult_Success
} TryReadResult;

TryReadResult ParseGameDataError(ParsedGameData *gameData, ParserState *state, const char *msg)
{
  gameData->parseError = TextFormat("Error at line %d: %s", ParserStateGetLineNumber(state), msg);
  return TryReadResult_Error;
}

TryReadResult ParseGameDataTryReadKeyInt(ParsedGameData *gameData, ParserState *state, const char *key, int *value, int minRange, int maxRange)
{
  if (!TextIsEqual(state->nextToken, key))
  {
    return TryReadResult_NoMatch;
  }

  if (!ParserStateReadNextInt(state, value))
  {
    return ParseGameDataError(gameData, state, TextFormat("Failed to read %s int value", key));
  }

  // range test, if minRange == maxRange, we don't check the range
  if (minRange != maxRange && (*value < minRange || *value > maxRange))
  {
    return ParseGameDataError(gameData, state, TextFormat(
      "Invalid value range for %s, range is [%d, %d], value is %d", 
      key, minRange, maxRange, *value));
  }

  return TryReadResult_Success;
}

TryReadResult ParseGameDataTryReadLevelSection(ParsedGameData *gameData, ParserState *state)
{
  int levelId;
  TryReadResult result = ParseGameDataTryReadKeyInt(gameData, state, "Level", &levelId, 0, 31);
  if (result != TryReadResult_Success)
  {
    return result;
  }
  
  gameData->lastLevelIndex = levelId;
  Level *level = &gameData->levels[levelId];

  // since we require the initialGold to be initialized with at least 1, we can use it as a flag
  // to detect if the level was already initialized
  if (level->initialGold != 0)
  {
    return ParseGameDataError(gameData, state, TextFormat("level %d already initialized", levelId));
  }

  int initialGoldInitialized = 0;

  while (1)
  {
    // try to read the next token and if we don't know how to handle it,
    // we rewind and return
    ParserState prevState = *state;
    
    if (!ParserStateReadNextToken(state))
    {
      // end of file
      break;
    }

    switch (ParseGameDataTryReadKeyInt(gameData, state, "initialGold", &level->initialGold, 1, 100000))
    {
      case TryReadResult_Success: 
        if (initialGoldInitialized)
        {
          return ParseGameDataError(gameData, state, "initialGold already initialized");
        }
        initialGoldInitialized = 1;
        continue;
      case TryReadResult_Error: return TryReadResult_Error;
    }

    // no match, return to previous state and break
    *state = prevState;
    break;
  }

  if (!initialGoldInitialized)
  {
    return ParseGameDataError(gameData, state, "initialGold not initialized");
  }

  printf("Parsed level %d, initialGold=%d\n", levelId, level->initialGold);
  return TryReadResult_Success;
}

int ParseGameData(ParsedGameData *gameData, ParserState *state)
{
  *gameData = (ParsedGameData){0};
  gameData->lastLevelIndex = -1;

  while (ParserStateReadNextToken(state))
  {
    switch (ParseGameDataTryReadLevelSection(gameData, state))
    {
      case TryReadResult_Success: continue;
      case TryReadResult_Error: return 0;
    }

    // read other sections later
  }

  return 1;
}

#define EXPECT(cond, msg) if (!(cond)) { printf("Error %s:%d: %s\n", __FILE__, __LINE__, msg); failedCount++; } else { passedCount++; }

int RunParseTests()
{
  int passedCount = 0, failedCount = 0;
  ParserState state;
  ParsedGameData gameData = {0};

  state = (ParserState) {"Level 7\n  initialGold 100\nLevel 2 initialGold 200", 0, {0}};
  gameData = (ParsedGameData) {0};
  EXPECT(ParseGameData(&gameData, &state) == 1, "Failed to parse level section");
  EXPECT(gameData.lastLevelIndex == 2, "lastLevelIndex should be 2");
  EXPECT(gameData.levels[7].initialGold == 100, "initialGold should be 100");
  EXPECT(gameData.levels[2].initialGold == 200, "initialGold should be 200");

  state = (ParserState) {"Level 392\n", 0, {0}};
  gameData = (ParsedGameData) {0};
  EXPECT(ParseGameData(&gameData, &state) == 0, "Expected to fail parsing level section");
  EXPECT(TextFindIndex(gameData.parseError, "Invalid value range for Level") >= 0, "Expected invalid value range error");
  EXPECT(TextFindIndex(gameData.parseError, "line 1") >= 0, "Expected to find line number 1");
  
  state = (ParserState) {"Level 3\n initialGold -34", 0, {0}};
  gameData = (ParsedGameData) {0};
  EXPECT(ParseGameData(&gameData, &state) == 0, "Expected to fail parsing level section");
  EXPECT(TextFindIndex(gameData.parseError, "Invalid value range for initialGold") >= 0, "Expected invalid value range error");
  EXPECT(TextFindIndex(gameData.parseError, "line 2") >= 0, "Expected to find line number 1");
  
  state = (ParserState) {"Level 3\n initialGold 2\n initialGold 3", 0, {0}};
  gameData = (ParsedGameData) {0};
  EXPECT(ParseGameData(&gameData, &state) == 0, "Expected to fail parsing level section");
  EXPECT(TextFindIndex(gameData.parseError, "line 3") >= 0, "Expected to find line number 1");

  state = (ParserState) {"Level 3", 0, {0}};
  gameData = (ParsedGameData) {0};
  EXPECT(ParseGameData(&gameData, &state) == 0, "Expected to fail parsing level section");

  state = (ParserState) {"Level 7\n  initialGold 100\nLevel 7 initialGold 200", 0, {0}};
  gameData = (ParsedGameData) {0};
  EXPECT(ParseGameData(&gameData, &state) == 0, "Expected to fail parsing level section");
  EXPECT(TextFindIndex(gameData.parseError, "line 3") >= 0, "Expected to find line number 1");

  
  printf("Passed %d test(s), Failed %d\n", passedCount, failedCount);

  return failedCount;
}

int main()
{
  printf("Running parse tests\n");
  if (RunParseTests())
  {
    return 1;
  }
  printf("\n");

  char *fileContent = LoadFileText("data/level.txt");
  if (fileContent == NULL)
  {
    printf("Failed to load file\n");
    return 1;
  }

  ParserState state = {fileContent, 0, {0}};
  ParsedGameData gameData = {0};

  if (!ParseGameData(&gameData, &state))
  {
    printf("Failed to parse game data: %s\n", gameData.parseError);
    UnloadFileText(fileContent);
    return 1;
  }

  UnloadFileText(fileContent);
  return 0; 
}
